
/*
  Program uChipTFT24.ino by G.Carrera for uChip rev. 04/01/2021
  this program uses the following components:
    a uChip card,
    a 2.4" touch TFT display (240x320 pixels)
    a BME280 sensor for temperature, pressure and humidity
  
  The FFT and BME libraries were written by Limor Fried/Ladyada for Adafruit Industries
  The XPT2046 touch library is written by Paul Stoffregen, paul@pjrc.com
  The uChip library is from www.itaca-innovation.com
*/
//include library code
#include <Adafruit_GFX.h>
#include <Adafruit_ILI9341.h>
#include <uChipPowerManagement.h>
#include <Wire.h>
#include <SPI.h>
#include <SD.h>
#include <Adafruit_Sensor.h>
#include <Adafruit_BME280.h>
#include <XPT2046_Touchscreen.h>

#define SEALEVELPRESSURE_HPA (1013.25)
Adafruit_BME280 bme; // I2C
#include <RTCZero.h>
RTCZero rtc;
//***********************************************************************************
//define pins of TFT screen
#define TFT_CS     7
#define TFT_DC     12
#define TFTLed  15
Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC);// hasware SPI
#define SD_CS  4 // CS SD card slot on display
//define pins of BME280
#define SCL  11
#define SDA  10
// pin for touchscreen
#define T_CS  2 
#define T_IRQ  13
XPT2046_Touchscreen ts(T_CS, T_IRQ);// interrupt enabled polling

boolean SDok = true;
bool sensorOK = true;
long days1900, secondspd;
unsigned int count= 0; // sample counter
float temperature, pressure, humidity;
int GBuff [3][240];// data buffer for graphics
int BuffInd= 0;// data buffer index
boolean first_buffer = true;
int ValMin [3] = {500, 11000, 1000};
int ValMax [3] = {-200, 1000, 0};
int GVal [3];
const char ValName[4] = {'T', 'P', 'H'};
const char* monthName[] = {"Gen", "Feb", "Mar", "Apr", "Mag", "Giu", "Lug", "Ago","Set", "Ott", "Nov", "Dic"};
byte ho,mi,se,da,mo,ye;
int period = 60;// sample period in seconds
boolean dateOk = false;
int X,Y;
long Number;
boolean ok = false;
boolean Acq = false;
unsigned long epoch, pepoch = 0;
int ng,nm;
char filename [13];
File dataFile;
//***********************************************************************************
void setup(void) {
  uChipSetVextValue(VEXT_3V3);
  delay(100); 
  analogReadResolution(12);
  pinMode(TFTLed, OUTPUT);// TFT backlight (ON = LOW)
  rtc.begin(); // initialize RTC
  ts.begin();// initialize XPT2046_Touchscreen
  Wire.begin (SDA, SCL);
  //if (!SD.begin(SD_CS)) SDok = false;   
  tft.begin();
  digitalWrite(TFTLed, LOW);// turn ON display leds
  tft.fillScreen(ILI9341_BLACK);
  tft.setTextColor(ILI9341_CYAN);
  tft.setTextSize(2);
  tft.setCursor(0,50);
  tft.print("uChip Meteo Logger");
  tft.setCursor(0,84);
  if (SD.begin(SD_CS)) tft.print("SD initialised!");
  else {
    SDok = false;
    tft.setTextColor(ILI9341_RED);
    tft.print("SD card failed!");    
  }
  if (!bme.begin(0x76, &Wire)) {  
    sensorOK = false;
    tft.setCursor(0, 42);
    tft.setTextColor(ILI9341_RED);
    tft.print("No BME280 sensor");
    while (1); // sketch halts in an endless loop
  }
  delay(5000);
  tft.fillScreen(ILI9341_BLACK);
  digitalWrite(TFTLed, HIGH);// turn OFF display leds
} // end of setup

void loop() {
  epoch = rtc.getEpoch();
  if(epoch - pepoch >= period){
    pepoch = epoch;  
    ReadDateTime(); // read the time from the DS3231 RTC chip
    days1900= long(epoch/86400)+ 25569;//days elapsed since 1/1/1900
    secondspd= ho*3600 + mi*60 + se;//seconds elapsed in the day
    tft.fillScreen(ILI9341_BLACK);
    digitalWrite(TFTLed, HIGH);// turn OFF display leds
    if (sensorOK){
      temperature = bme.readTemperature();
      pressure = bme.readPressure()/100.0F;// pressure in hPascal
      humidity = bme.readHumidity();
      GVal[0] = int(temperature*10);//stores temperature in tenths of a degree
      GVal[1] = int(pressure*10);//stores pressure in tenths of a millibar
      GVal[2] = int(humidity)*10;//stores relative humidity in per thousand
      // find the max and min of measurements
      for (int j=0; j <= 2; j++){
        if (GVal[j] > ValMax[j]) ValMax[j] = GVal[j];
        if (GVal[j] < ValMin[j]) ValMin[j] = GVal[j];
      }
      if (first_buffer){// store them in the buffer 
        for (int j=0; j <= 2; j++){
          GBuff[j][BuffInd]= GVal[j];
        }  
        BuffInd++;
        if (BuffInd == 240) first_buffer = false;      
      } else {
         for (int j=0; j <= 2; j++){
           for (int i=0; i <= 238; i++){
             GBuff[j][i]= GBuff [j][i+1];// shift left previous values
           }
         }
         for (int j=0; j <= 2; j++){
           GBuff[j][239]= GVal[j];// update buffer with last values
         }  
      }
      if (SDok && Acq) SaveMeasure();// save measurements if SD is present and acquisition started
    }
  }
    if (ts.tirqTouched()) {//has the display been touched?
      if (ts.touched()) {
        digitalWrite(TFTLed, LOW);
        do {
          drawMainMenu();// draws the menu keys
          DetectCommands();// interprets and executes the command         
        } while (!ok);
      }
    }
}
//*********** FUNCTIONS ****************************************************
void ReadDateTime() {
  ho= rtc.getHours();
  mi= rtc.getMinutes();
  se= rtc.getSeconds();
  da= rtc.getDay();
  mo= rtc.getMonth();  
  ye= rtc.getYear();
}
void DisplayDateTime() {
  tft.setTextColor(ILI9341_GREEN);
  tft.setTextSize(2);
  tft.setCursor(0,12);
  tft.print("Data : ");
  print2digits(da);
  tft.print(" ");
  tft.print(monthName[mo-1]);
  tft.print(" ");
  tft.print(ye+2000);
  tft.setCursor(0,30);
  tft.print("Ore : ");
  print2digits(ho);
  tft.print(":");
  print2digits(mi);
  tft.print(":");
  print2digits(se);
  tft.setCursor(0,48);
  tft.print("Tc = ");
  tft.print(period);
}

void print2digits(int n) {
  if (n < 10) tft.print("0");
  tft.print(n);
}

void DisplayMeasures(){
  tft.setTextColor(ILI9341_YELLOW);
  tft.setTextSize(2); 
  tft.setCursor(0, 66);
  tft.print("T [C]= ");
  tft.print(temperature,1);
  tft.setCursor(0, 84);
  tft.setTextColor(ILI9341_CYAN);
  tft.print("P [hPa]= ");
  tft.print(pressure,1);
  tft.setCursor(0, 102);
  tft.setTextColor(ILI9341_WHITE);
  tft.print("RH [%]= ");
  tft.print(humidity,0);
  if (Acq){
    tft.setCursor(0, 120);
    tft.setTextColor(ILI9341_YELLOW);
    tft.print("Nc= ");
    tft.print(count);
    tft.setCursor(0,192);
    tft.print("File:");
    tft.print(filename);     
  }
  if (BuffInd > 3){ //print the statistics only after a few samples
    tft.setTextColor(ILI9341_WHITE);   
    for (int m=0; m <= 2; m++){
      tft.setCursor(0, 138+m*18);
      tft.print(ValName[m]);
      tft.print(" = ");   
      tft.print(ValMin[m]/10.0,1);
      tft.print(" -> ");
      tft.print(ValMax[m]/10.0,1);
    }
  }
}

String Conv4digits(int n) {
  if (n < 10) return "000" + String(n);
  else if (n < 100) return "00" + String(n);
  else if (n < 1000) return "0" + String(n);
  else return String(n);
}

void SaveMeasure(){
  dataFile = SD.open(filename, FILE_WRITE);
  // if the file is available and SD is Ok, write to it:
  if (dataFile) {
    String dataString = String(days1900)+','+String(secondspd)+',';// days since 1900, seconds per day,
    dataString += String(temperature,1)+','+String(pressure,1)+','+String(humidity,0);
    dataFile.println(dataString);
    dataFile.close();
    count++;
   } else {
    SDok = false;
    digitalWrite(TFTLed, LOW);// turn ON display leds
    tft.fillScreen(ILI9341_BLACK);
    tft.setCursor(0, 101);
    tft.setTextColor(ILI9341_RED);
    tft.print("FILE ERROR!");
    delay(2000);
   }
}

void PlotMeasures(int k) {
 //plot k graph points of all measures
  const int yp [3] = {2,106,210};// upper left corners of the plots 
  digitalWrite(TFTLed, LOW);// turn ON display leds
  tft.fillScreen(ILI9341_BLACK);
  tft.setTextColor(ILI9341_WHITE);
  tft.setTextSize(2);
  for (int m=0; m <= 2; m++){
    tft.drawRect(0,yp[m],240,100,ILI9341_WHITE);
    tft.setCursor(10, yp[m]+5);
    tft.print(ValName[m]);
    for (int j=0; j <= k; j++){
      int val = map(GBuff[m][j], ValMin[m], ValMax[m], 99, 0);
      tft.drawPixel(j, val+yp[m], ILI9341_YELLOW);// draw the graph
    }
  }
}

unsigned int GetNum(String prompt){
// print a numeric keyboard and a prompt message
// and get a numeric unsigned integer
 tft.fillScreen(ILI9341_BLACK);
 Number=0;
 draw_Keypad();// draw the virtual keyboard
 tft.setCursor(10, 20);
 tft.setTextSize(2);
 tft.setTextColor(ILI9341_YELLOW);
 tft.println(prompt); //print prompt
 do { // repeat until ok is pressed
  if (ts.touched()) {
    TS_Point p = ts.getPoint();
    X= p.y; Y= p.x;// invert the axes
    DetectKeys();
  }
 }while (!ok);
 ok = false;
 return Number;
 }

void draw_Keypad() {
  String symbol[3][4] = {
    { "6", "7", "8", "9" },
    { "2", "3", "4", "5" },
    { "C", "0", "1", "ok" }
  };
  //Draw the Result Box
  tft.fillRect(0, 60, 240, 80,ILI9341_CYAN);
 //Draw keys
  tft.fillRect(0,260,60,60,ILI9341_RED);
  tft.fillRect(0,140,240,120,ILI9341_BLUE);
  tft.fillRect(60,260,120,60,ILI9341_BLUE);
  tft.fillRect(180,260,60,60,ILI9341_GREEN);
  //Draw Horizontal Lines
  for (int h=140; h<=320; h+=60)
  tft.drawFastHLine(0, h, 240,ILI9341_WHITE);
  //Draw Vertical Lines
  for (int v=0; v<=240; v+=60)
  tft.drawFastVLine(v, 140, 1800,ILI9341_WHITE);
  //Display keypad lables 
  for (int j=0;j<3;j++) {
    for (int i=0;i<4;i++) {
      tft.setCursor(22 + (60*i), 160 + (60*j));
      tft.setTextSize(3);
      tft.setTextColor(ILI9341_WHITE);
      if (j==2 && i==3) tft.setTextColor(ILI9341_BLACK);// ok is more visible
      tft.println(symbol[j][i]);
    }
  }
}

void drawMainMenu() {
  String Commands [3] = {"MISURE","GRAFICI","SETTAGGI"};
  tft.fillScreen(ILI9341_BLACK);
  // draw keys
  tft.fillRect(0,0,240,80,ILI9341_BLACK);
  tft.setTextSize(2);
  tft.setTextColor(ILI9341_YELLOW);
  tft.setCursor(10, 30);
  tft.println("SCEGLIERE OPZIONE:");
  tft.fillRoundRect(0,80,240,60,15,ILI9341_PINK);// ACQ START/STOP
  tft.fillRoundRect(0,140,240,60,15,ILI9341_LIGHTGREY);// "MISURE"
  tft.fillRoundRect(0,200,240,60,15,ILI9341_YELLOW);//"GRAFICI" 
  tft.fillRoundRect(0,260,240,60,15,ILI9341_CYAN);// "SETTAGGI"
  tft.drawRoundRect(0,80,240,60,15,ILI9341_BLACK);
  tft.drawRoundRect(0,140,240,60,15,ILI9341_BLACK);
  tft.drawRoundRect(0,200,240,60,15,ILI9341_BLACK);
  tft.drawRoundRect(0,260,240,60,15,ILI9341_BLACK);
  tft.setTextColor(ILI9341_BLACK);
  tft.setCursor(20, 105);
  tft.setTextSize(2);
  if (!Acq && SDok){
    tft.println("Acq. Start/Stop");    
  } else if (SDok) tft.println("Acquisizione..."); 
  tft.setTextSize(3);  
  for (int j=0;j<4;j++) {
     tft.setCursor(50, 160 + (60*j));
     tft.println(Commands[j]);
  }
}

void DetectCommands(){
  ok = false;
  unsigned long epin = rtc.getEpoch();
  do {   
    if (ts.touched()) {
      TS_Point p = ts.getPoint();
      X= p.y; Y= p.x;// invert the axes
      if (Y>1040 && Y<1719){// ACQ START/STOP
         if (!Acq && SDok){
           Acq = true;
           count= 0;
           tft.fillRoundRect(0,81,240,59,15,ILI9341_RED);// ACQ START/STOP
           tft.setTextColor(ILI9341_WHITE);
           tft.setCursor(10, 105);
           tft.setTextSize(2);
           tft.println("Inizio Acquisizione");
           // get the file name from the timestamp
           ng= int((epoch-1604188800)/86400);// days since 01 Nov 2020 00:00:00
           if (ng < 0) ng = 0;// check the limit of days
           if (ng > 9999) ng= 9999;// days must be 4 characters long
           nm= int(secondspd/10);// seconds/10 from midnight
           String fn = Conv4digits(ng)+ Conv4digits(nm)+ ".TXT";// file name
           fn.toCharArray(filename,13);// String to char array conversion
           delay(1000);
         } else if (SDok){
           Acq = false;
           tft.fillRoundRect(0,81,240,59,15,ILI9341_GREEN);// ACQ START/STOP
           tft.setTextColor(ILI9341_BLUE);
           tft.setCursor(20, 105);
           tft.setTextSize(2);
           tft.println("Fine Acquisizione");
           delay(1000);
           ok = true;
         }           
      } else if (Y>1720 && Y<2449){// "MISURE"
        tft.fillScreen(ILI9341_BLACK);
        digitalWrite(TFTLed, LOW);
        DisplayDateTime();
        DisplayMeasures();
        ok = true;
      } else if (Y>2450 && Y<3099){// "GRAFICI"
        if (first_buffer) PlotMeasures(BuffInd);
        else PlotMeasures(239);        
        ok = true;
      } else if (Y>3100 && Y<3780){ // "SETTAGGI"
	    if (!Acq ){// does not modify the parameters if it is acquiring
          Settings();
          ok = true;
		}
      }
    }  
  if (rtc.getEpoch()- epin >= 5) ok = true;// exit after 5 seconds
  }while (!ok);
}

void DetectKeys(){
  if (X<3800 && X>3027){ //Detecting Buttons on Column 1
    if (Y>3100 && Y<3780) Number=0; //If cancel Button is pressed
    if (Y>2450 && Y<3100){ //If Button 2 is pressed
      if (Number==0) Number= 2;
      else Number = (Number*10) + 2; //Pressed twice
    }  
    if (Y>1720 && Y<2450){ //If Button 6 is pressed
      if (Number==0) Number=6;
      else Number = (Number*10) + 6; //Pressed twice
    }
  }
  if (X<3020 && X>2100){ //Detecting Buttons on Column 2 
    if (Y>3100 && Y<3740){ //Button 0 is Pressed
      if (Number==0) Number=0;
      else Number = (Number*10) + 0; //Pressed twice
    }
    if (Y>2450 && Y<3110){ //Button 3 is Pressed
      if (Number==0) Number=3;
      else Number = (Number*10) + 3; //Pressed twice
    }
    if (Y>1770 && Y<2485){ //Button 7 is Pressed
      if (Number==0) Number=7;
      else Number = (Number*10) + 7; //Pressed twic
    }
  }
  if (X<2100 && X>1140){ //Detecting Buttons on Column 3
    if (Y>3146 && Y<3750){ //Button 1 is Pressed
       if (Number==0) Number=1;
       else Number = (Number*10) + 1; //Pressed twice
    }  
    if (Y>2430 && Y<3146){ //Button 4 is Pressed
       if (Number==0) Number=4;
       else Number = (Number*10) + 4; //Pressed twice
    }  
    if (Y>1760 && Y<2432){ //Button 8 is Pressed
       if (Number==0) Number=8;
       else Number = (Number*10) + 8; //Pressed twice
    }
  }
  if (X<1150 && X>290){ //Detecting Buttons on Column 4 
    if (Y>3129 && Y<3780) ok = true; //Button ok is Pressed   
    if (Y>2442 && Y<3129){ //Button 5 is Pressed
       if (Number==0) Number=5;
       else Number = (Number*10) + 5; //Pressed twice
    }
    if (Y>1767 && Y<2441){ //Button 9 is Pressed
       if (Number==0) Number=9;
       else Number = (Number*10) + 9; //Pressed twice
    }
  }
  tft.fillRect(0, 60, 240, 80, ILI9341_CYAN);  //clear result box
  tft.setCursor(10, 80);
  tft.setTextSize(4);
  tft.setTextColor(ILI9341_BLACK);
  tft.println(Number); //update new value
  delay(100);
}
void Settings(){ // set the sample period and RTC clock
  period = GetNum("Periodo [s] = ");// Sample Period
  ye = GetNum("Anno (20xx) = ");
  mo = GetNum("Mese = ");
  da = GetNum("Giorno = ");
  ho = GetNum("Ore = ");
  mi = GetNum("Minuti = ");
  se = GetNum("Secondi = ");
  rtc.setTime(ho, mi, se);// update the TRC clock
  rtc.setDate(da, mo, ye);
}
